!+
! Module: imrebin_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: The task rebins images.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module imrebin_mod

  use dal
  use param  
  use errorHandling

  use dal_aux
  use array_utils
!!!  use wcs_aux
use l_wcs_aux

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
    inImageSetName,&
    outImageSetName
  integer(int16) ::&
    binXSize,&
    binYSize,&
    offsetX,&
    offsetY
  character(8) ::&
    newValueStyle

  ! Other globals:
  type(WcsT) :: inWcs, outWcs

  public imrebin

contains 

  !---------------------------------------------------------------------
  subroutine imrebin
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
    integer(int32) ::&
      inDataType, outDataType, inXSize, inYSize, outXSize, outYSize,&
      backstepX, backstepY, xi, yi, inXiStart, inXiFinis,&
      inYiStart, inYiFinis, numPixels
    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld
    real(double), pointer :: inImage(:,:)
    real(double), allocatable :: outImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    inSet = dataSet(inImageSetName, READ)
    inArray = array(inSet, 0)
    inDataType = dataType(inArray)
    call readArrayData(inArray, inImage) ! in ssclib/dal_aux
    inXSize = size(inImage,1)
    inYSize = size(inImage,2)

    ! Decide new image size:

    backstepX = modulo((binXSize - offsetX), binXSize)
    backstepY = modulo((binYSize - offsetY), binYSize)
    outXSize = ceiling((inXSize + backstepX)/real(binXSize))
    outYSize = ceiling((inYSize + backstepY)/real(binYSize))

    allocate(outImage(outXSize, outYSize))

    do xi = 1, outXSize
      inXiStart = max(1, (xi - 1) * binXSize - backstepX    )
      inXiFinis = min(inXSize, inXiStart + binXSize - 1)

      do yi = 1, outYSize
        inYiStart = max(1, (yi - 1) * binYSize - backstepY    )
        inYiFinis = min(inYSize, inYiStart + binYSize - 1)

        select case(newValueStyle)
        case('sum')
          outImage(xi, yi)&
            = sum(inImage(inXiStart:inXiFinis,inYiStart:inYiFinis))

        case('average')
          numPixels = (1 + inXiFinis - inXiStart) * (1 + inYiFinis - inYiStart)

          outImage(xi, yi)&
            = sum(inImage(inXiStart:inXiFinis,inYiStart:inYiFinis))&
            / real(numPixels, double)

        case('rms')
          numPixels = (1 + inXiFinis - inXiStart) * (1 + inYiFinis - inYiStart)

          outImage(xi, yi)&
            = sqrt(sum(inImage(inXiStart:inXiFinis,inYiStart:inYiFinis)&
            * inImage(inXiStart:inXiFinis,inYiStart:inYiFinis))&
            / real(numPixels, double))
        case default
call error('badNewValueStyle',"The value of the --newvaluestyle parameter &
&was not recognized.")
        end select
      end do
    end do

    ! Decide new image data type:

    if (newValueStyle .ne. 'sum') then
      if (maxval(abs(outImage)) < huge(1.0_single)) then
        outDataType = REAL32
      else
        outDataType = REAL64
      end if
    else
      outDataType = minDataType(maxval(abs(outImage)), inDataType)
    end if

    ! Write the output data to file:

    outSet = dataSet(outImageSetName, CREATE)
    outArray = addArray(outSet, 'PRIMARY', outDataType, (/outXSize,outYSize/))
    call loadOutArray(outImage, outArray, outDataType)
    call setAttributes(attributable(outArray), attributable(inArray))

    ! Fix up new wcs:

    call readWcs(inSet, inArray, inWcs) ! in ssclib/wcs_aux
    call release(inSet)
    outWcs = inWcs

    call convertNewToOld(inWcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)
    ! in ssclib/wcs_aux
    xPixelSizeWorld = xPixelSizeWorld / real(binXSize)
    yPixelSizeWorld = yPixelSizeWorld / real(binYSize)
    call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, outWcs)
    ! in ssclib/wcs_aux

    outWcs%x%refPixels = (inWcs%x%refPixels - 0.5 - offsetX)&
      / real(binXSize) + 0.5
    if (offsetX > 0) outWcs%x%refPixels = outWcs%x%refPixels + 1.0
    outWcs%y%refPixels = (inWcs%y%refPixels - 0.5 - offsetY)&
      / real(binYSize) + 0.5
    if (offsetY > 0) outWcs%y%refPixels = outWcs%y%refPixels + 1.0

    call writeWcs(outArray, outWcs) ! in ssclib/wcs_aux
    call release(outSet)

    deallocate(inImage)

  end subroutine imrebin

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    inImageSetName = stringParameter('inset')
    outImageSetName = stringParameter('outset')

    binXSize = intParameter('binxsize')
    binYSize = intParameter('binysize')
    offsetX  = intParameter('offsetx')
    offsetY  = intParameter('offsety')

    if (offsetX >= binXSize) then
      call warning('offsetXTooBig',"")
    end if
    if (offsetY >= binYSize) then
      call warning('offsetYTooBig',"")
    end if

    offsetX = modulo(offsetX, binXSize)
    offsetY = modulo(offsetY, binYSize)

    newValueStyle = stringParameter('newvaluestyle')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine loadOutArray(outImage, outArray, outDataType)
  !---------------------------------------------------------------------
!**** put in dal_aux?

    implicit none

    real(double),   intent(in) :: outImage(:,:)
    type(ArrayT),   intent(in) :: outArray
    integer(int32), intent(in) :: outDataType

    integer(int8),  pointer :: outDataInt8(:,:)
    integer(int16), pointer :: outDataInt16(:,:)
    integer(int32), pointer :: outDataInt32(:,:)
    real(single),   pointer :: outDataReal32(:,:)
    real(double),   pointer :: outDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start

    select case(outDataType)
    case(INTEGER8)
      outDataInt8 => int8Array2Data(outArray)
      outDataInt8 = nint(outImage, int8)

    case(INTEGER16)
      outDataInt16 => int16Array2Data(outArray)
      outDataInt16 = nint(outImage, int16)

    case(INTEGER32)
      outDataInt32 => int32Array2Data(outArray)
      outDataInt32 = nint(outImage, int32)

    case(REAL32)
      outDataReal32 => real32Array2Data(outArray)
      outDataReal32 = real(outImage, single)

    case(REAL64)
      outDataReal64 => real64Array2Data(outArray)
      outDataReal64 = outImage

!    case(BOOLEAN)
!      inDataBool => boolArray2Data(outArray)

    case default
call error('badArrayDataType',"")
    end select

  end subroutine loadOutArray

  !---------------------------------------------------------------------
  integer(int32) function minDataType(maxValue, dataTypeFloor)
  !---------------------------------------------------------------------
!**** put this in dal_aux

    implicit none

    real(double), intent(in) :: maxValue
    integer(int32), intent(in) :: dataTypeFloor

    integer(int32) :: dataTypes(5), dataTypeFloorIndex

    !-------------------------------------------------------------------
    ! Start

    dataTypes = (/INTEGER8, INTEGER16, INTEGER32, REAL32, REAL64/)

if (count(dataTypes == dataTypeFloor) .ne. 1) call error('',"")
    dataTypeFloorIndex&
      = smallestTrueIndex(logical(dataTypes==dataTypeFloor,bool))
    ! in ssclib/array_utils.

    ! First see if dataTypeFloor can contain the max value:

    select case(dataTypeFloor)
    case(INTEGER8)
      if (maxValue < huge(1_INT8)) then
        minDataType = dataTypeFloor
  return
      end if

    case(INTEGER16)
      if (maxValue < huge(1_INT16)) then
        minDataType = dataTypeFloor
  return
      end if

    case(INTEGER32)
      if (maxValue < huge(1_INT32)) then
        minDataType = dataTypeFloor
  return
      end if

    case(REAL32)
      if (maxValue < huge(1.0_single)) then
        minDataType = dataTypeFloor
  return
      end if

    case(REAL64)
      minDataType = dataTypeFloor
  return

    case default
call error('',"")
    end select

    ! If not, have to have a larger data type:

    if (maxValue < huge(1_INT16)) then
      minDataType = INTEGER16
  return
    end if

    if (maxValue < huge(1_INT32)) then
      minDataType = INTEGER32
  return
    end if

    if (maxValue < huge(1.0_single)) then
      minDataType = REAL32
  return
    end if

    minDataType = REAL64

  end function minDataType

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module imrebin_mod
