!This originates from ssclib-3.7.
!
! Module: l_wcs_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Routines to facilitate handling of wcs values.
! 
! Comments: TODO:
!	As of 2004-01-30 the library generally supports the WCS system outlined in Greisen and Calabretta A&A 395, 1061-1075 (2002). However, a couple of caveats: (i) only the RA---TAN/DEC--TAN projection is supported; (ii) if CD1_1 is present, all CDi_j kwds are required to be present; (iii) PCi_j kwds are not yet supported.
!
! $Id: l_wcs_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module l_wcs_aux

  use dal
  use errorHandling	! used in dumpWcs and the pixels<->tan routines.

  use constants
  use dal_aux	! contains subs getRealKwd(), getRealKwd(), getStringKwd().
  use ssc_misc	! for approxEqual
  use array_utils	! for vectorCross

  implicit none 

  private 

  public initializeWcsT
  public copyWcsT

  public hasWcs
  public readWcs
  public writeWcs
  public deleteWcs
  public dumpWcs

  public easifyWcsType
  public wcsAreEquivalent
  public wcsPixelAreaWorld
  public checkImageEquivalence

  public pixelsToTan
  public tanToPixels

public readWcsSourceList
interface readWcsSourceList
  module procedure readWcsSourceListSet, readWcsSourceListName
end interface
!*** this is necessary until src lists conform to ogip.

  public convertOldToNew
  public convertNewToOld

  interface hasWcs
    module procedure hasWcsSet, hasWcsArray, hasWcsTab
  end interface

  interface readWcs
    module procedure readWcsColumnsName, readWcsColumns
    module procedure readWcsSetArray, readWcsArray
    module procedure readWcsSet, readWcsSetName, readWcsSetArrayName
    module procedure readWcsTable, readWcsSetArrayNumber
  end interface

  interface writeWcs
    module procedure writeWcsColumns, writeWcsArray
  end interface

  interface easifyWcsType
    module procedure easifyWcsTypeScalar
    module procedure easifyWcsTypeVector
  end interface

  interface tanToPixels
    module procedure tanToPixelsScalar, tanToPixelsVector, tanToPixelsArray
  end interface

  interface pixelsToTan
    module procedure pixelsToTanScalar, pixelsToTanVector
  end interface

  type, public :: WcsAxisT
    character(len=8) ::&
      ctype,&
      worldUnit
    real(double) ::&
      matrixOnAxis,&
      matrixOffAxis,&
      refPixels,&
      refWorld
  end type WcsAxisT

  type, public :: WcsT
    type(WcsAxisT) :: x, y
    logical(bool) ::&
      crotFound
  end type WcsT

logical :: useNewWcsStyle = .false.

contains 

  !---------------------------------------------------------------------
  logical(bool) function hasWcsSet(set)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set

    !-------------------------------------------------------------------
    ! Start

    hasWcsSet = .false. ! default

    if (    hasAttribute(set, 'CRPIX1')&
    .and.   hasAttribute(set, 'CRPIX2')&
    .and.   hasAttribute(set, 'CRVAL1')&
    .and.   hasAttribute(set, 'CRVAL2')&
    .and.   hasAttribute(set, 'CTYPE1')&
    .and.   hasAttribute(set, 'CTYPE2')&
    .and. ((hasAttribute(set, 'CDELT1')&
    .and.   hasAttribute(set, 'CDELT2'))&
    .or.   (hasAttribute(set, 'CD1_1')&
    .and.   hasAttribute(set, 'CD1_2')&
    .and.   hasAttribute(set, 'CD2_1')&
    .and.   hasAttribute(set, 'CD2_2'))))&
      hasWcsSet = .true.

  end function hasWcsSet

  !---------------------------------------------------------------------
  logical(bool) function hasWcsArray(inArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray

    !-------------------------------------------------------------------
    ! Start

    hasWcsArray = .false. ! default

    if (    hasAttribute(inArray, 'CRPIX1')&
    .and.   hasAttribute(inArray, 'CRPIX2')&
    .and.   hasAttribute(inArray, 'CRVAL1')&
    .and.   hasAttribute(inArray, 'CRVAL2')&
    .and.   hasAttribute(inArray, 'CTYPE1')&
    .and.   hasAttribute(inArray, 'CTYPE2')&
    .and. ((hasAttribute(inArray, 'CDELT1')&
    .and.   hasAttribute(inArray, 'CDELT2'))&
    .or.   (hasAttribute(inArray, 'CD1_1')&
    .and.   hasAttribute(inArray, 'CD1_2')&
    .and.   hasAttribute(inArray, 'CD2_1')&
    .and.   hasAttribute(inArray, 'CD2_2'))))&
      hasWcsArray = .true.

  end function hasWcsArray

  !---------------------------------------------------------------------
  logical(bool) function hasWcsTab(tab)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab

    !-------------------------------------------------------------------
    ! Start

    hasWcsTab = .false. ! default

    if (    hasAttribute(tab, 'CRPIX1')&
    .and.   hasAttribute(tab, 'CRPIX2')&
    .and.   hasAttribute(tab, 'CRVAL1')&
    .and.   hasAttribute(tab, 'CRVAL2')&
    .and.   hasAttribute(tab, 'CTYPE1')&
    .and.   hasAttribute(tab, 'CTYPE2')&
    .and. ((hasAttribute(tab, 'CDELT1')&
    .and.   hasAttribute(tab, 'CDELT2'))&
    .or.   (hasAttribute(tab, 'CD1_1')&
    .and.   hasAttribute(tab, 'CD1_2')&
    .and.   hasAttribute(tab, 'CD2_1')&
    .and.   hasAttribute(tab, 'CD2_2'))))&
      hasWcsTab = .true.

  end function hasWcsTab

  !---------------------------------------------------------------------
  subroutine initializeWcsT(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start

    wcs%x%ctype          = ''
    wcs%x%worldUnit      = ''
    wcs%x%matrixOnAxis   = 1.0d0
    wcs%x%matrixOffAxis  = 0.0d0
    wcs%x%refPixels      = 0.0d0
    wcs%x%refWorld       = 0.0d0

    wcs%y%ctype          = ''
    wcs%y%worldUnit      = ''
    wcs%y%matrixOnAxis   = 1.0d0
    wcs%y%matrixOffAxis  = 0.0d0
    wcs%y%refPixels      = 0.0d0
    wcs%y%refWorld       = 0.0d0

    wcs%crotFound        = .false.

  end subroutine initializeWcsT

  !---------------------------------------------------------------------
  function copyWcsT(inWcs) result(outWcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: inWcs
    type(WcsT)             :: outWcs

    !-------------------------------------------------------------------
    ! Start

    outWcs%x%ctype          = inWcs%x%ctype
    outWcs%x%worldUnit      = inWcs%x%worldUnit
    outWcs%x%matrixOnAxis   = inWcs%x%matrixOnAxis
    outWcs%x%matrixOffAxis  = inWcs%x%matrixOffAxis
    outWcs%x%refPixels      = inWcs%x%refPixels
    outWcs%x%refWorld       = inWcs%x%refWorld

    outWcs%y%ctype          = inWcs%y%ctype
    outWcs%y%worldUnit      = inWcs%y%worldUnit
    outWcs%y%matrixOnAxis   = inWcs%y%matrixOnAxis
    outWcs%y%matrixOffAxis  = inWcs%y%matrixOffAxis
    outWcs%y%refPixels      = inWcs%y%refPixels
    outWcs%y%refWorld       = inWcs%y%refWorld

    outWcs%crotFound      = inWcs%crotFound

  end function copyWcsT

  !---------------------------------------------------------------------
  subroutine readWcsColumnsName(tab, xColName, yColName, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: xColName, yColName
    type(WcsT),   intent(out) :: wcs

    type(ColumnT) :: xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    xCol = column(tab, xColName)
    yCol = column(tab, yColName)

    call readWcsColumns(xCol, yCol, wcs)

    call release(xCol)
    call release(yCol)

  end subroutine readWcsColumnsName

  !---------------------------------------------------------------------
  subroutine readWcsColumns(xCol, yCol, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol
    type(WcsT),   intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype          = stringAttribute(xCol, 'TCTYP')
    wcs%y%ctype          = stringAttribute(yCol, 'TCTYP')

    call checkWcsType(wcs)

    wcs%x%refPixels      = real64Attribute(xCol, 'TCRPX')
    wcs%y%refPixels      = real64Attribute(yCol, 'TCRPX')
    wcs%x%refWorld       = real64Attribute(xCol, 'TCRVL')
    wcs%y%refWorld       = real64Attribute(yCol, 'TCRVL')

    wcs%crotFound = .false. ! default
    if (hasAttribute(xCol, "TCD1_")) then
      wcs%x%matrixOnAxis  = real64Attribute(xCol, 'TCD1_')
      wcs%x%matrixOffAxis = real64Attribute(xCol, 'TCD2_')
!      wcs%y%matrixOnAxis  = real64Attribute(yCol, 'TCD1_')
!      wcs%y%matrixOffAxis = real64Attribute(yCol, 'TCD2_')
wcs%y%matrixOffAxis = real64Attribute(yCol, 'TCD1_')
wcs%y%matrixOnAxis  = real64Attribute(yCol, 'TCD2_')

    else
      xPixelSizeWorld = real64Attribute(xCol, 'TCDLT')
      yPixelSizeWorld = real64Attribute(yCol, 'TCDLT')
      if (hasAttribute(yCol, "TCROT")) then
        rotAngle = real64Attribute(yCol, 'TCROT')
        wcs%crotFound = .true.

      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(xCol, "TCUNI")) then
      wcs%x%worldUnit = stringAttribute(xCol, 'TCUNI')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(yCol, "TCUNI")) then
      wcs%y%worldUnit = stringAttribute(yCol, 'TCUNI')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsColumns

  !---------------------------------------------------------------------
  subroutine checkWcsType(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    if (.not. (wcs%x%ctype .eq. 'RA---TAN' .and. wcs%y%ctype .eq. 'DEC--TAN'))&
call error('notSupported',"")

  end subroutine checkWcsType

  !---------------------------------------------------------------------
  subroutine getDefaultXWorldUnit(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(inout) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    select case(wcs%x%ctype)
    case('RA---TAN')
      wcs%x%worldUnit = 'deg'
    case default
call error('notSupported',"")
    end select

  end subroutine getDefaultXWorldUnit

  !---------------------------------------------------------------------
  subroutine getDefaultYWorldUnit(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(inout) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    select case(wcs%y%ctype)
    case('DEC--TAN')
      wcs%y%worldUnit = 'deg'
    case default
call error('notSupported',"")
    end select

  end subroutine getDefaultYWorldUnit

  !---------------------------------------------------------------------
  subroutine readWcsSetName(setName, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*), intent(in)  :: setName
    type(WcsT),   intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, 0, READ)
    call readWcsSetArray(set, inArray, wcs)

    call release(set)

  end subroutine readWcsSetName

  !---------------------------------------------------------------------
  subroutine readWcsSetArrayName(setName, arrayName, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*), intent(in)  :: setName, arrayName
    type(WcsT),   intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayName, READ)
    call readWcsSetArray(set, inArray, wcs)

    call release(set)

  end subroutine readWcsSetArrayName

  !---------------------------------------------------------------------
  subroutine readWcsSetArrayNumber(setName, arrayNumber, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*),   intent(in)  :: setName
    integer(int32), intent(in)  :: arrayNumber
    type(WcsT),     intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayNumber, READ)
    call readWcsSetArray(set, inArray, wcs)

    call release(set)

  end subroutine readWcsSetArrayNumber

  !---------------------------------------------------------------------
  subroutine readWcsSetArray(set, inArray, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    type(DataSetT), intent(in)  :: set
    type(ArrayT),   intent(in)  :: inArray
    type(WcsT),     intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld
    integer :: status

    !-------------------------------------------------------------------
    ! Start:

    call getStringKwd("CTYPE1", set, inArray, wcs%x%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%y%ctype)

    call checkWcsType(wcs)

    call getRealKwd("CRPIX1",   set, inArray, wcs%x%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%y%refPixels)
    call getRealKwd("CRVAL1",   set, inArray, wcs%x%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%y%refWorld)

    wcs%crotFound = .false. ! default
    call getRealKwd("CD1_1", set, inArray, wcs%x%matrixOnAxis, status)
    if (status == 0) then ! the keyword was found, get the others:
      call getRealKwd("CD2_1", set, inArray, wcs%x%matrixOffAxis)
!      call getRealKwd("CD1_2", set, inArray, wcs%y%matrixOnAxis)
!      call getRealKwd("CD2_2", set, inArray, wcs%y%matrixOffAxis)
call getRealKwd("CD1_2", set, inArray, wcs%y%matrixOffAxis)
call getRealKwd("CD2_2", set, inArray, wcs%y%matrixOnAxis)

    else
      call getRealKwd("CDELT1", set, inArray, xPixelSizeWorld)
      call getRealKwd("CDELT2", set, inArray, yPixelSizeWorld)

      call getRealKwd("CROTA2", set, inArray, rotAngle, status)
      if (status == 0) then ! the CROTA keyword was found:
        wcs%crotFound = .true.

      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    call getStringKwd("CUNIT1", set, inArray, wcs%x%worldUnit, status)
    if (status .ne. 0) then
      call getDefaultXWorldUnit(wcs)
    end if

    call getStringKwd("CUNIT2", set, inArray, wcs%y%worldUnit, status)
    if (status .ne. 0) then
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsSetArray

  !---------------------------------------------------------------------
  subroutine readWcsArray(inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in)  :: inArray
    type(WcsT),   intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(inArray, 'CTYPE1')
    wcs%y%ctype     = stringAttribute(inArray, 'CTYPE2')

    call checkWcsType(wcs)

    wcs%x%refPixels = real64Attribute(inArray, 'CRPIX1')
    wcs%y%refPixels = real64Attribute(inArray, 'CRPIX2')
    wcs%x%refWorld  = real64Attribute(inArray, 'CRVAL1')
    wcs%y%refWorld  = real64Attribute(inArray, 'CRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(inArray, "CD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(inArray, 'CD1_1')
      wcs%x%matrixOffAxis = real64Attribute(inArray, 'CD2_1')
!      wcs%y%matrixOnAxis  = real64Attribute(inArray, 'CD1_2')
!      wcs%y%matrixOffAxis = real64Attribute(inArray, 'CD2_2')
wcs%y%matrixOffAxis = real64Attribute(inArray, 'CD1_2')
wcs%y%matrixOnAxis  = real64Attribute(inArray, 'CD2_2')

    else
      xPixelSizeWorld = real64Attribute(inArray, 'CDELT1')
      yPixelSizeWorld = real64Attribute(inArray, 'CDELT2')
      if (hasAttribute(inArray, "CROTA2")) then
        rotAngle = real64Attribute(inArray, 'CROTA2')
        wcs%crotFound = .true.

      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(inArray, "CUNIT1")) then
      wcs%x%worldUnit = stringAttribute(inArray, 'CUNIT1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(inArray, "CUNIT2")) then
      wcs%y%worldUnit = stringAttribute(inArray, 'CUNIT2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsArray

  !---------------------------------------------------------------------
  subroutine readWcsSet(set, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in)  :: set
    type(WcsT),     intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(set, 'CTYPE1')
    wcs%y%ctype     = stringAttribute(set, 'CTYPE2')

    call checkWcsType(wcs)

    wcs%x%refPixels = real64Attribute(set, 'CRPIX1')
    wcs%y%refPixels = real64Attribute(set, 'CRPIX2')
    wcs%x%refWorld  = real64Attribute(set, 'CRVAL1')
    wcs%y%refWorld  = real64Attribute(set, 'CRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(set, "CD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(set, 'CD1_1')
      wcs%x%matrixOffAxis = real64Attribute(set, 'CD2_1')
!      wcs%y%matrixOnAxis  = real64Attribute(set, 'CD1_2')
!      wcs%y%matrixOffAxis = real64Attribute(set, 'CD2_2')
wcs%y%matrixOffAxis = real64Attribute(set, 'CD1_2')
wcs%y%matrixOnAxis  = real64Attribute(set, 'CD2_2')

    else
      xPixelSizeWorld = real64Attribute(set, 'CDELT1')
      yPixelSizeWorld = real64Attribute(set, 'CDELT2')
      if (hasAttribute(set, "CROTA2")) then
        rotAngle = real64Attribute(set, 'CROTA2')
        wcs%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(set, "CUNIT1")) then
      wcs%x%worldUnit = stringAttribute(set, 'CUNIT1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(set, "CUNIT2")) then
      wcs%y%worldUnit = stringAttribute(set, 'CUNIT2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsSet

  !---------------------------------------------------------------------
  subroutine readWcsTable(tab, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    type(WcsT),   intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(tab, 'CTYPE1')
    wcs%y%ctype     = stringAttribute(tab, 'CTYPE2')

    call checkWcsType(wcs)

    wcs%x%refPixels = real64Attribute(tab, 'CRPIX1')
    wcs%y%refPixels = real64Attribute(tab, 'CRPIX2')
    wcs%x%refWorld  = real64Attribute(tab, 'CRVAL1')
    wcs%y%refWorld  = real64Attribute(tab, 'CRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(tab, "CD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(tab, 'CD1_1')
      wcs%x%matrixOffAxis = real64Attribute(tab, 'CD2_1')
!      wcs%y%matrixOnAxis  = real64Attribute(tab, 'CD1_2')
!      wcs%y%matrixOffAxis = real64Attribute(tab, 'CD2_2')
wcs%y%matrixOffAxis = real64Attribute(tab, 'CD1_2')
wcs%y%matrixOnAxis  = real64Attribute(tab, 'CD2_2')

    else
      xPixelSizeWorld = real64Attribute(tab, 'CDELT1')
      yPixelSizeWorld = real64Attribute(tab, 'CDELT2')
      if (hasAttribute(tab, "CROTA2")) then
        rotAngle = real64Attribute(tab, 'CROTA2')
        wcs%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(tab, "CUNIT1")) then
      wcs%x%worldUnit = stringAttribute(tab, 'CUNIT1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(tab, "CUNIT2")) then
      wcs%y%worldUnit = stringAttribute(tab, 'CUNIT2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsTable

  !---------------------------------------------------------------------
  subroutine readWcsSourceListSet(set, wcs)
  !---------------------------------------------------------------------
!*** These kwds should surely be stored as proper column kwds?

    implicit none

    type(DataSetT), intent(in)  :: set
    type(WcsT),     intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(set, 'TCTYPE1')
    wcs%y%ctype     = stringAttribute(set, 'TCTYPE2')

    call checkWcsType(wcs)

    wcs%x%refPixels = real64Attribute(set, 'TCRPIX1')
    wcs%y%refPixels = real64Attribute(set, 'TCRPIX2')
    wcs%x%refWorld  = real64Attribute(set, 'TCRVAL1')
    wcs%y%refWorld  = real64Attribute(set, 'TCRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(set, "TCD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(set, 'TCD1_1')
      wcs%x%matrixOffAxis = real64Attribute(set, 'TCD2_1')
!      wcs%y%matrixOnAxis  = real64Attribute(set, 'TCD1_2')
!      wcs%y%matrixOffAxis = real64Attribute(set, 'TCD2_2')
wcs%y%matrixOffAxis = real64Attribute(set, 'TCD1_2')
wcs%y%matrixOnAxis  = real64Attribute(set, 'TCD2_2')

    else
      xPixelSizeWorld = real64Attribute(set, 'TCDELT1')
      yPixelSizeWorld = real64Attribute(set, 'TCDELT2')
      if (hasAttribute(set, "TCROTA2")) then
        rotAngle = real64Attribute(set, 'TCROTA2')
        wcs%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(set, "TCUNI1")) then
      wcs%x%worldUnit = stringAttribute(set, 'TCUNI1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(set, "TCUNI2")) then
      wcs%y%worldUnit = stringAttribute(set, 'TCUNI2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsSourceListSet

  !---------------------------------------------------------------------
  subroutine readWcsSourceListName(setName, wcs)
  !---------------------------------------------------------------------
!*** These kwds should surely be stored as proper column kwds?

    implicit none

    character(*), intent(in) :: setName
    type(WcsT), intent(out) :: wcs

    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    call readWcsSourceListSet(set, wcs)
    call release(set)

  end subroutine readWcsSourceListName

  !---------------------------------------------------------------------
  subroutine writeWcsColumns(xCol, yCol, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol
    type(WcsT),    intent(in) :: wcs

    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(xCol, "TCTYP", wcs%X%ctype)
    call setAttribute(xCol, "TCRPX", wcs%X%refPixels)
    call setAttribute(xCol, "TCRVL", wcs%X%refWorld)
    call setAttribute(xCol, 'TCUNI', wcs%X%worldUnit)

    call setAttribute(yCol, "TCTYP", wcs%Y%ctype)
    call setAttribute(yCol, "TCRPX", wcs%Y%refPixels)
    call setAttribute(yCol, "TCRVL", wcs%Y%refWorld)
    call setAttribute(yCol, 'TCUNI', wcs%Y%worldUnit)

    if (useNewWcsStyle) then
      call setAttribute(xCol, 'TCD1_',  wcs%X%matrixOnAxis)
      call setAttribute(xCol, 'TCD2_',  wcs%X%matrixOffAxis)

!      call setAttribute(yCol, 'TCD1_',  wcs%Y%matrixOnAxis)
!      call setAttribute(yCol, 'TCD2_',  wcs%Y%matrixOffAxis)
call setAttribute(yCol, 'TCD1_',  wcs%Y%matrixOffAxis)
call setAttribute(yCol, 'TCD2_',  wcs%Y%matrixOnAxis)

    else ! use the old style.
      call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

      call setAttribute(xCol, "TCDLT", xPixelSizeWorld)
      call setAttribute(yCol, "TCDLT", yPixelSizeWorld)
      call setAttribute(yCol, "TCROT", rotAngle)
    end if

  end subroutine writeWcsColumns

  !---------------------------------------------------------------------
  subroutine writeWcsArray(inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    type(WcsT),   intent(in) :: wcs

    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(inArray, 'CRPIX1', wcs%X%refPixels)
    call setAttribute(inArray, 'CRPIX2', wcs%Y%refPixels)
    call setAttribute(inArray, 'CRVAL1', wcs%X%refWorld)
    call setAttribute(inArray, 'CRVAL2', wcs%Y%refWorld)
    call setAttribute(inArray, 'CTYPE1', wcs%X%ctype)
    call setAttribute(inArray, 'CTYPE2', wcs%Y%ctype)

    call setAttribute(inArray, 'CUNIT1', wcs%X%worldUnit)
    call setAttribute(inArray, 'CUNIT2', wcs%Y%worldUnit)

    if (useNewWcsStyle) then
      call setAttribute(inArray, 'CD1_1',  wcs%X%matrixOnAxis)
      call setAttribute(inArray, 'CD2_1',  wcs%X%matrixOffAxis)
!      call setAttribute(inArray, 'CD1_2',  wcs%Y%matrixOnAxis)
!      call setAttribute(inArray, 'CD2_2',  wcs%Y%matrixOffAxis)
call setAttribute(inArray, 'CD1_2',  wcs%Y%matrixOffAxis)
call setAttribute(inArray, 'CD2_2',  wcs%Y%matrixOnAxis)

    else ! use the old style.
      call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

      call setAttribute(inArray, 'CROTA2', rotAngle)
      call setAttribute(inArray, 'CDELT1', xPixelSizeWorld)
      call setAttribute(inArray, 'CDELT2', yPixelSizeWorld)
    end if

  end subroutine writeWcsArray

  !---------------------------------------------------------------------
  subroutine deleteWcs(set)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set

    integer, parameter :: numRootStrs = 6, maxNumDims = 3
    character(5) :: rootStrs(numRootStrs)
    integer :: i, j
    character(6) :: str1
    character(5) :: str2

    !-------------------------------------------------------------------
    ! Start

    rootStrs = (/'CRPIX', 'CRVAL', 'CTYPE', 'CDELT', 'CROTA', 'CUNIT'/)

    do i = 1, maxNumDims
      do j = 1, numRootStrs
        write(str1, '(a5, i1)') rootStrs(j), i
        if (hasAttribute(set, str1)) call deleteAttribute(set, str1)
      end do
      do j = 1, maxNumDims
        write(str2, '(a2, i1, a1, i1)') 'CD', i, '_', j
        if (hasAttribute(set, str2)) call deleteAttribute(set, str2)
        write(str2, '(a2, i1, a1, i1)') 'PC', i, '_', j
        if (hasAttribute(set, str2)) call deleteAttribute(set, str2)
      end do
    end do

  end subroutine deleteWcs

  !---------------------------------------------------------------------
  subroutine easifyWcsTypeScalar(ctypeX, ctypeY, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctypeX, ctypeY
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctypeX) .eq. 'DETX' .and. trim(ctypeY) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctypeX).eq.'RA---TAN' .and. trim(ctypeY) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsTypeScalar

  !---------------------------------------------------------------------
  subroutine easifyWcsTypeVector(ctype, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctype(2)
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctype(1)) .eq. 'DETX' .and. trim(ctype(2)) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctype(1)).eq.'RA---TAN' .and. trim(ctype(2)) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsTypeVector

  !---------------------------------------------------------------------
  subroutine dumpWcs(wcs, verbosityArg)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: wcs
    integer(int32), intent(in), optional :: verbosityArg

    integer(int32) :: verbosity

    !-------------------------------------------------------------------
    ! Start

    if (present(verbosityArg)) then
      verbosity = verbosityArg
    else
      verbosity = VERBOSE
    end if

    call message(AppMsg, verbosity, 'WCS dump:')
    call message(AppMsg, verbosity, '  X axis:')
    call message(AppMsg, verbosity, '    CTYPE: '//wcs%x%ctype)
    call message(AppMsg, verbosity, '    CUNIT: '//wcs%x%worldUnit)
    write(errstr, '(a, es23.16)')   '    CRVAL:', wcs%x%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CRPIX:', wcs%x%refPixels
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CD1_1:', wcs%x%matrixOnAxis
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CD2_1:', wcs%x%matrixOffAxis
    call message(AppMsg, verbosity)
!    write(errstr, '(a, es23.16)')   '    CDELT:', wcs%x%pixelSizeWorld
!    call message(AppMsg, verbosity)

    call message(AppMsg, verbosity, '  Y axis:')
    call message(AppMsg, verbosity, '    CTYPE: '//wcs%y%ctype)
    call message(AppMsg, verbosity, '    CUNIT: '//wcs%y%worldUnit)
    write(errstr, '(a, es23.16)')   '    CRVAL:', wcs%y%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CRPIX:', wcs%y%refPixels
    call message(AppMsg, verbosity)
!    write(errstr, '(a, es23.16)')   '    CD1_2:', wcs%y%matrixOnAxis
write(errstr, '(a, es23.16)')   '    CD1_2:', wcs%y%matrixOffAxis
    call message(AppMsg, verbosity)
!    write(errstr, '(a, es23.16)')   '    CD2_2:', wcs%y%matrixOffAxis
write(errstr, '(a, es23.16)')   '    CD2_2:', wcs%y%matrixOnAxis
    call message(AppMsg, verbosity)
!    write(errstr, '(a, es23.16)')   '    CDELT:', wcs%y%pixelSizeWorld
!    call message(AppMsg, verbosity)
!**** fix this

  end subroutine dumpWcs

  !---------------------------------------------------------------------
  subroutine pixelsToTanScalar(xPixels, yPixels, wcs, tanX, tanY)
  !---------------------------------------------------------------------
  ! See pixelsToTanVector for a description.

    implicit none

    real(single), intent(in)  :: xPixels, yPixels
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: tanX, tanY

    real(single) :: xPixelsVec(1), yPixelsVec(1)
    real(double) :: tanXVec(1), tanYVec(1)

    !-------------------------------------------------------------------
    ! Start

!    tanX = (xPixels - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
!    tanY = (yPixels - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180

    xPixelsVec(1) = xPixels
    yPixelsVec(1) = yPixels

    call pixelsToTanVector(xPixelsVec, yPixelsVec, wcs, tanXVec, tanYVec)

    tanX = tanXVec(1)
    tanY = tanYVec(1)

  end subroutine pixelsToTanScalar

  !---------------------------------------------------------------------
  subroutine pixelsToTanVector(xPixels, yPixels, wcs, tanX, tanY)
  !---------------------------------------------------------------------
  ! This transform is only valid for CTYPE1='RA---TAN' and CTYPE2='DEC--TAN'.
  !
  ! The tan system coordinates are aligned with the ra/dec world coordinate
  ! system, but are expressed in natural units rather than in whatever units
  ! are specified for the world coordinates. In addition, the tanX axis
  ! points to the right (in the direction of decreasing ra) and the tanY
  ! axis points upwards (in the direction of increasing dec) when the view
  ! is from the centre of the projection sphere outwards.
  !
  ! Numerical examples:
  !
  !	wcs%x%ctype = 'RA---TAN'
  !	wcs%y%ctype = 'DEC--TAN'
  !	wcs%x%refWorld = 25 degrees ra
  !	wcs%y%refWorld = 0 degrees dec
  !
  !	In this case, a position with ra = 30 deg and dec = 0 would give
  !	tanX = -tan(30-25) = -0.0875, tanY = 0; whereas a position with
  !	ra = 0 deg and dec = 7 would give tanX = 0, tanY = tan(7) = 0.123.
  !
  ! As described in Greisen & Calabretta A&A 395, 1061 (2002) equation (3),
  ! the relation between pixel coordinates p_i and intermediate world
  ! coordinates x_i is
  !
  !	      _ N
  !	x_i = >_  CDi_j (p_j - CRPIXj).
  !	     j = 1
  !
  ! This is implemented in the present subroutine, then the result x_i
  ! is converted to my tan system.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: xPixels(:), yPixels(size(xPixels))
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: tanX(size(xPixels)), tanY(size(xPixels))

    real(double) ::&
      intermediateWorldXToTanX, intermediateWorldYToTanY,&
      intermediateWorldX(size(xPixels)), intermediateWorldY(size(xPixels))
    real(single) :: relPixelsX(size(xPixels)), relPixelsY(size(xPixels))

    !-------------------------------------------------------------------
    ! Start

!    tanX = (xPixels - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
!    tanY = (yPixels - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180

!    if (wcs%x%ctype .ne. 'RA---TAN' .or. wcs%y%ctype .ne. 'DEC--TAN')&
!call error('',"")

    call checkWcsType(wcs)

    select case(wcs%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('notSupported',"")
    end select

    select case(wcs%y%worldUnit)
    case('deg')
      intermediateWorldYToTanY = PiOn180
    case default
call error('notSupported',"")
    end select

    relPixelsX = xPixels - wcs%x%refPixels
    relPixelsY = yPixels - wcs%y%refPixels

    intermediateWorldX = wcs%x%matrixOnAxis  * relPixelsX&
                       + wcs%y%matrixOffAxis * relPixelsY
    intermediateWorldY = wcs%y%matrixOnAxis  * relPixelsY&
                       + wcs%x%matrixOffAxis * relPixelsX

    tanX = intermediateWorldX * intermediateWorldXToTanX
    tanY = intermediateWorldY * intermediateWorldYToTanY

  end subroutine pixelsToTanVector

  !---------------------------------------------------------------------
  subroutine tanToPixelsScalar(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! See tanToPixelsVector for a description.

    implicit none

    real(double), intent(in)  :: tanX, tanY
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: xPixels, yPixels

    real(single) :: xPixelsVec(1), yPixelsVec(1)
    real(double) :: tanXVec(1), tanYVec(1)

    !-------------------------------------------------------------------
    ! Start

!    xPixels = tanX / abs(wcs%x%pixelSizeWorld) / PiOn180 + wcs%x%refPixels 
!    yPixels = tanY / abs(wcs%y%pixelSizeWorld) / PiOn180 + wcs%y%refPixels 

    tanXVec(1) = tanX
    tanYVec(1) = tanY

    call tanToPixelsVector(wcs, tanXVec, tanYVec, xPixelsVec, yPixelsVec)

    xPixels = xPixelsVec(1)
    yPixels = yPixelsVec(1)

  end subroutine tanToPixelsScalar

  !---------------------------------------------------------------------
  subroutine tanToPixelsVector(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! This transform is only valid for CTYPE1='RA---TAN' and CTYPE2='DEC--TAN'.
  !
  ! The tan system coordinates are aligned with the ra/dec world coordinate
  ! system, but are expressed in natural units rather than in whatever units
  ! are specified for the world coordinates. In addition, the tanX axis
  ! points to the right (in the direction of decreasing ra) and the tanY
  ! axis points upwards (in the direction of increasing dec) when the view
  ! is from the centre of the projection sphere outwards.
  !
  ! Numerical examples:
  !
  !	wcs%x%ctype = 'RA---TAN'
  !	wcs%y%ctype = 'DEC--TAN'
  !	wcs%x%refWorld = 25 degrees ra
  !	wcs%y%refWorld = 0 degrees dec
  !
  !	In this case, a position with ra = 30 deg and dec = 0 would give
  !	tanX = -tan(30-25) = -0.0875, tanY = 0; whereas a position with
  !	ra = 0 deg and dec = 7 would give tanX = 0, tanY = tan(7) = 0.123.
  !
  ! As described in Greisen & Calabretta A&A 395, 1061 (2002) equation (3),
  ! the relation between pixel coordinates p_i and intermediate world
  ! coordinates x_i is
  !
  !	      _ N
  !	x_i = >_  CDi_j (p_j - CRPIXj).
  !	     j = 1
  !
  ! This is inverted in the present subroutine, after x_i is first converted
  ! from my tan system.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: tanX(:), tanY(size(tanX))
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: xPixels(size(tanX)), yPixels(size(tanX))

    real(double) ::&
      intermediateWorldXToTanX, intermediateWorldYToTanY,&
      intermediateWorldX(size(xPixels)), intermediateWorldY(size(xPixels)),&
      det, invMatrix(2,2)
    real(single) :: relPixelsX(size(xPixels)), relPixelsY(size(xPixels))

    !-------------------------------------------------------------------
    ! Start

!    xPixels = tanX / abs(wcs%x%pixelSizeWorld) / PiOn180 + wcs%x%refPixels 
!    yPixels = tanY / abs(wcs%y%pixelSizeWorld) / PiOn180 + wcs%y%refPixels 

    if (wcs%x%ctype .ne. 'RA---TAN' .or. wcs%y%ctype .ne. 'DEC--TAN')&
call error('notSupported',"")

    select case(wcs%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('notSupported',"")
    end select

    select case(wcs%y%worldUnit)
    case('deg')
      intermediateWorldYToTanY = PiOn180
    case default
call error('notSupported',"")
    end select

    intermediateWorldX = tanX / intermediateWorldXToTanX
    intermediateWorldY = tanY / intermediateWorldYToTanY

    det = wcs%x%matrixOnAxis  * wcs%y%matrixOnAxis&
        - wcs%x%matrixOffAxis * wcs%y%matrixOffAxis
if (det == 0.0) call error('singularWcsRotMatrix',"")

    invMatrix(1,1) =  wcs%y%matrixOnAxis  / det		! 1st index row,
    invMatrix(1,2) = -wcs%y%matrixOffAxis / det		! 2nd index col.
    invMatrix(2,1) = -wcs%x%matrixOffAxis / det
    invMatrix(2,2) =  wcs%x%matrixOnAxis  / det

    relPixelsX = invMatrix(1,1) * intermediateWorldX&
               + invMatrix(1,2) * intermediateWorldY
    relPixelsY = invMatrix(2,1) * intermediateWorldX&
               + invMatrix(2,2) * intermediateWorldY

    xPixels = relPixelsX + wcs%x%refPixels
    yPixels = relPixelsY + wcs%y%refPixels

  end subroutine tanToPixelsVector

  !---------------------------------------------------------------------
  subroutine tanToPixelsArray(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! See tanToPixelsVector for a description.

    implicit none

    real(double), intent(in)  :: tanX(:,:),&
                                 tanY(size(tanX,1),size(tanX,2))
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: xPixels(size(tanX,1),size(tanX,2)),&
                                 yPixels(size(tanX,1),size(tanX,2))

    real(single) ::&
      xPixelsVec(size(tanX,1)*size(tanX,2)),&
      yPixelsVec(size(tanX,1)*size(tanX,2))
    real(double) ::&
      tanXVec(size(tanX,1)*size(tanX,2)), tanYVec(size(tanX,1)*size(tanX,2))
    integer(int32) :: xSize, ySize, vSize

    !-------------------------------------------------------------------
    ! Start

!    xPixels = tanX / abs(wcs%x%pixelSizeWorld) / PiOn180 + wcs%x%refPixels 
!    yPixels = tanY / abs(wcs%y%pixelSizeWorld) / PiOn180 + wcs%y%refPixels 

    xSize = size(tanX, 1)
    ySize = size(tanX, 2)
    vSize = xSize * ySize

    tanXVec = reshape(tanX, (/vSize/))
    tanYVec = reshape(tanY, (/vSize/))

    call tanToPixelsVector(wcs, tanXVec, tanYVec, xPixelsVec, yPixelsVec)

    xPixels = reshape(xPixelsVec, (/xSize, ySize/))
    yPixels = reshape(yPixelsVec, (/xSize, ySize/))

  end subroutine tanToPixelsArray

  !---------------------------------------------------------------------
  logical(bool) function wcsAreEquivalent(wcs1, wcs2)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: wcs1, wcs2

    !-------------------------------------------------------------------
    ! Start

    wcsAreEquivalent = .false. ! default

  if (wcs1%x%ctype          .ne. wcs2%x%ctype         ) return
  if (.not. approxEqual(wcs1%x%refPixels,      wcs2%x%refPixels)     ) return
  if (.not. approxEqual(wcs1%x%refWorld,       wcs2%x%refWorld)      ) return

  if (wcs1%y%ctype          .ne. wcs2%y%ctype         ) return
  if (.not. approxEqual(wcs1%y%refPixels,      wcs2%y%refPixels)     ) return
  if (.not. approxEqual(wcs1%y%refWorld,       wcs2%y%refWorld)      ) return

!*** fix up

    wcsAreEquivalent = .true.

  end function wcsAreEquivalent

  !---------------------------------------------------------------------
  subroutine convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
  !---------------------------------------------------------------------
  ! Makes use of equation (189) from Calabretta & Greisen A&A 395 1077
  ! (2002), ie
  !
  !	(CD1_1 CD1_2)   (CDELT1*cos[CROTA2]  -CDELT2*sin[CROTA2])
  !	(           ) = (                                       ).
  !	(CD2_1 CD2_2)   (CDELT1*sin[CROTA2]   CDELT2*cos[CROTA2])
  !
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)    :: xPixelSizeWorld, yPixelSizeWorld, rotAngle
    type(WcsT),   intent(inout) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%matrixOnAxis   =  xPixelSizeWorld * cos(rotAngle)
    wcs%x%matrixOffAxis  =  xPixelSizeWorld * sin(rotAngle)
    wcs%y%matrixOffAxis  = -yPixelSizeWorld * sin(rotAngle)
    wcs%y%matrixOnAxis   =  yPixelSizeWorld * cos(rotAngle)

  end subroutine convertOldToNew

  !---------------------------------------------------------------------
  subroutine convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)
  !---------------------------------------------------------------------
  ! Makes use of equation (189) from Calabretta & Greisen A&A 395 1077
  ! (2002), ie
  !
  !	(CD1_1 CD1_2)   (CDELT1*cos[CROTA2]  -CDELT2*sin[CROTA2])
  !	(           ) = (                                       ).
  !	(CD2_1 CD2_2)   (CDELT1*sin[CROTA2]   CDELT2*cos[CROTA2])
  !
  !---------------------------------------------------------------------

    implicit none

    type(WcsT),   intent(in) :: wcs
    real(double), intent(out) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    real(double) :: rotAngleAlt, rotAngle2

    !-------------------------------------------------------------------
    ! Start:

    xPixelSizeWorld = sqrt(wcs%X%matrixOnAxis * wcs%X%matrixOnAxis&
      + wcs%X%matrixOffAxis * wcs%X%matrixOffAxis)

    yPixelSizeWorld = sqrt(wcs%Y%matrixOnAxis * wcs%Y%matrixOnAxis&
      + wcs%Y%matrixOffAxis * wcs%Y%matrixOffAxis)

    rotAngle    = atan2(-wcs%Y%matrixOffAxis, wcs%Y%matrixOnAxis)
    rotAngleAlt = atan2( wcs%X%matrixOffAxis, wcs%X%matrixOnAxis)

    if (abs(rotAngleAlt - rotAngle) > PiOn2) then ! one of the pixel
    ! dimensions should be negative. It is immaterial which, but I'll
    ! pick the x-axis dimension as recommended by C&G:
      xPixelSizeWorld = -xPixelSizeWorld

      ! Just as a check:

      rotAngle2 = atan2(-wcs%X%matrixOffAxis, -wcs%X%matrixOnAxis)
      if (abs(rotAngle2 - rotAngle) > PiOn2)&
call error('ssclibWcsBug',"This should not be able to happen. Please &
&contact the code developer.")
    end if

  end subroutine convertNewToOld

  !---------------------------------------------------------------------
  function wcsPixelAreaWorld(wcs) result(area)
  !---------------------------------------------------------------------
  ! Calculates the area in world coordinates of an image pixel.

    implicit none

    type(WcsT), intent(in) :: wcs
    real(double)           :: area

    real(single) :: squareXPixels(2,2), squareYPixels(2,2)
    real(double) ::&
      relPixelsX(2,2), relPixelsY(2,2),&
      parlogramXWorld(2,2), parlogramYWorld(2,2),&
      vectorA(3), vectorB(3), crossProd(3)

    !-------------------------------------------------------------------
    ! Start

    squareXPixels(:,2) = (/0.0, 1.0/)
    squareXPixels(:,1) = (/0.0, 1.0/)

    squareYPixels(:,2) = (/1.0, 1.0/)
    squareYPixels(:,1) = (/0.0, 0.0/)

    relPixelsX = squareXPixels - wcs%x%refPixels
    relPixelsY = squareYPixels - wcs%y%refPixels

    parlogramXWorld = wcs%x%matrixOnAxis  * relPixelsX&
                    + wcs%y%matrixOffAxis * relPixelsY
    parlogramYWorld = wcs%y%matrixOnAxis  * relPixelsY&
                    + wcs%x%matrixOffAxis * relPixelsX

    ! The transformed square has now (in general) been sheared into a
    ! parallelogram, then rotated and offset. We'll now load the sides
    ! of the transformed square into two vectors. Note that we need a z
    ! component in order to be able to do a cross product; but of course
    ! since all the above operations take place in the plane, z components
    ! are set to zero.

    vectorA = (/parlogramXWorld(2,1) - parlogramXWorld(1,1)&
              , parlogramYWorld(2,1) - parlogramYWorld(1,1), 0.0d0/)

    vectorB = (/parlogramXWorld(1,2) - parlogramXWorld(1,1)&
              , parlogramYWorld(1,2) - parlogramYWorld(1,1), 0.0d0/)

    ! Area of a parallelogram defined by two vectors is just the magnitude
    ! of their cross product:

    crossProd = vectorCross(vectorA, vectorB)
    area = sqrt(sum(crossProd * crossProd))

  end function wcsPixelAreaWorld

  !---------------------------------------------------------------------
  subroutine checkImageEquivalence(imageSetNames, equivalent, xSize, ySize, wcs)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in)  :: imageSetNames(:)
    logical(bool),  intent(out) :: equivalent
    integer(int16), intent(out) :: xSize, ySize
    type(WcsT),     intent(out) :: wcs

    integer(int16) :: numImages, imageNum, tempXSize, tempYSize

    !-------------------------------------------------------------------
    ! Start

    equivalent = .true. ! default

    xSize = 0 ! Don't like to leave
    ySize = 0 ! them unset.

    call initializeWcsT(wcs)

    numImages = size(imageSetNames)
  if (numImages <= 0) return

call readWcs(imageSetNames(1), wcs)
!*** should also read the wcs from each of the other arrays and test them for equivalence. But there are still problems with the wcs equiv test.

    call getArraySize(imageSetNames(1), xSize, ySize) ! in ssclib/dal_aux
    do imageNum = 2, numImages
      call getArraySize(imageSetNames(imageNum), tempXSize, tempYSize)
      if (tempXSize .ne. xSize .or. tempYSize .ne. ySize) then
        equivalent = .false.
  return
      end if
    end do

  end subroutine checkImageEquivalence

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module l_wcs_aux
