//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	orbitdump --- DUMP contents of ODF ORBIT file to stdout
//
//	DESCRIPTION:
//		The tasks dumps the contents of the orbit file in the current ODF
//		(pointed at via environment variable SAS_ODF) to stdout in the
//		format
//			t x y z d vx vy vz vabs
//
//		where
//			t			: time in Mission-Elapsed-Time [s]
//			x, y, z		: components of S/C position vector in Earth-centered
//						  J2000 reference frame [km]
//			d			: norm of vector (x, y, z) [km]
//			vx, vy, vz	: components of S/C velocity vector in Earth-centered
//						  J2000 reference frame [km/s]
//			vabs		: norm of vector (vx, vy, vz)
//
//		The time increment between successive elements is given via task
//		parameter <tstep> with d/f 10s. With <offsett>=true causes all time
//		tags to be offset by time of first element which, hence, is assigned
//		t=0.
//
//	USAGE:
//		orbitdump [tstep=<value>] [offsett=true|false] >orbit.dat
//
//	CHANGE HISTORY:
//		UL, 2002-06-14: 1.0.0: - initial version
//
//	AUTHOR:
//		U. Lammers, Fri Jun 14 10:25:54 MEST 2002
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/tools/src/orbitdump/orbitdump.cc,v 1.2 2002/06/14 10:24:18 ulammers Exp $
//
#include <cmath>
#include <iostream>
#include <iomanip>

#include "OalServer.h"
#include "XmmPositionLocator.h"
#include "Xmm.h"
#include "param.h"
#include "STime.h"

#include "orbitdump.h"

void orbitdump()
{
	const XmmPositionLocator *posLoc = oalServer.positionLocator();

	STime tFrom = posLoc->dataBegin(),
		  tTo = posLoc->dataEnd();
	double deltaT = realParameter("tstep");		// [s] 
	unsigned N = UNT((tTo-tFrom)/deltaT);
	bool offset = booleanParameter("offsett");

	double t0 = tFrom - Xmm::missionReferenceTime;
	for (unsigned n=0; n<N; ++n) {
		StateVector sv;
		double t = t0+DBL(n)*deltaT;
		try {
			sv = posLoc->stateVector(t);
		} catch(XmmPositionLocator::NoPositionDataAvailable) {
			continue;
		}
		CartesianVector &p = sv.position,
						 v = sv.velocity;
		cout<<setprecision(16)<<(offset ? t-t0 : t)<<' '
			<<setprecision(8)<<p.x()<<' '<<p.y()<<' '<<p.z()<<' '
			<<p.norm()<<' '
			<<setprecision(8)<<v.x()<<' '<<v.y()<<' '<<v.z()<<' '
			<<v.norm()<<endl;
	}
} // orbitdump()

